@@ -159,7 +159,7 @@ int init(int argc, char **argv, bcf_hdr_t *in, bcf_hdr_t *out)
         {"unrelated",1,0,'u'},
         {0,0,0,0}
     };
-    char c;
+    int c;
     while ((c = getopt_long(argc, argv, "?ht:u:p:",loptions,NULL)) >= 0)
     {
         switch (c) 
@@ -456,6 +456,7 @@ void flush_viterbi(args_t *args)
         s3 = bcf_hdr_int2id(args->hdr,BCF_DT_SAMPLE,args->ifather);
         s2 = bcf_hdr_int2id(args->hdr,BCF_DT_SAMPLE,args->ichild);
     }
+    else abort();
 
     if ( !args->fp )
     {
